@@ -6,6 +6,8 @@ from functools import partial
 from itertools import chain
 from operator import itemgetter
 
+from typing.re import Match
+
 from pynlple.processing.token import EMOJI_MODIFIERS_EXTENDED
 from .emojis import UnicodeClusteredEmojiProviderV11
 from .token import TOKEN_TYPES, DELIMITER_TYPES, DELIMITER_CONTEXTS, \
@@ -13,6 +15,8 @@ from .token import TOKEN_TYPES, DELIMITER_TYPES, DELIMITER_CONTEXTS, \
 
 URL_REGEX = r'((ht|f)tps?:[\\/]+|www\d{0,3}\.)([.…]+|[^\s\\/$.?#].[^\s]*)'
 
+BOLD_TAG_PATTERN = re.compile(r'</?b>')
+
 
 class IPreprocessor(object):
     """Interface to collect modules for text string preprocessing."""
@@ -177,43 +181,105 @@ class RegexReplacerAdapter(RegexReplacer):
         super().__init__(regex_query, replace_tag_with, case_sensitive, use_locale, restrict_to_ascii)
 
 
-def _get_bold_replacement(repl_):
-    stripd_repl_ = repl_.strip()
-    if len(stripd_repl_) == 0:
-        return repl_, repl_, repl_, repl_
-    else:
-        ind_ = repl_.index(stripd_repl_)
-        ind_end_ = ind_ + len(stripd_repl_)
-        return (repl_,
-                ''.join([repl_[:ind_], '<b>', repl_[ind_:]]),
-                ''.join([repl_[:ind_end_], '</b>', repl_[ind_end_:]]),
-                ''.join([repl_[:ind_], '<b>', repl_[ind_:ind_end_], '</b>', repl_[ind_end_:]]))
+class MatchPostprocessor(object):
+
+    def __call__(self, match: Match) -> str:
+        raise NotImplementedError()
 
 
-bold_tag_pattern = re.compile(r'</?b>')
+class PreserveHighlight(MatchPostprocessor):
 
+    def __init__(
+        self,
+        replace_with: str = "",
+    ):
+        self.replace_with = replace_with
+        self.__stripped_replace_with = self.replace_with.strip()
+        self.__start_index = self.replace_with.index(self.__stripped_replace_with)
+        self.__end_index = self.__start_index + len(self.__stripped_replace_with)
+        self.left_bound = ''.join([
+            self.replace_with[:self.__start_index],
+            '<b>',
+            self.replace_with[self.__start_index:]
+        ])
+        self.right_bound = ''.join([
+            self.replace_with[:self.__end_index],
+            '</b>',
+            self.replace_with[self.__end_index:]
+        ])
+        self.left_right_bound = ''.join([
+            self.replace_with[:self.__start_index],
+            '<b>',
+            self.replace_with[self.__start_index:self.__end_index],
+            '</b>',
+            self.replace_with[self.__end_index:]
+        ])
 
-def replace_w_bold_tag_preserved(repl_, left_bound_repl_, right_bound_repl_, left_right_bound_repl_, match):
-    match_text = match.group(0)
-    if '<b>' in match_text or '</b>' in match_text:
-        cnt = 0
-        for match in bold_tag_pattern.finditer(match_text):
-            if match.group(0) == '</b>':
-                cnt -= 1
+    def __call__(self, match: Match) -> str:
+        match_text = match.group(0)
+
+        if '<b>' in match_text or '</b>' in match_text:
+            cnt = 0
+            for match in BOLD_TAG_PATTERN.finditer(match_text):
+                if match.group(0) == '</b>':
+                    cnt -= 1
+                else:
+                    cnt += 1
+            if cnt > 0:
+                return self.left_bound
+            elif cnt < 0:
+                return self.right_bound
             else:
-                cnt += 1
-        if cnt > 0:
-            return left_bound_repl_
-        elif cnt < 0:
-            return right_bound_repl_
+                return self.left_right_bound
         else:
-            return left_right_bound_repl_
-    else:
-        return repl_
+            return self.replace_with
 
 
-def _get_target_func(repl_):
-    return partial(replace_w_bold_tag_preserved, *_get_bold_replacement(repl_))
+class PreserveHighlightedEntity(MatchPostprocessor):
+
+    def __init__(
+        self,
+        replace_with: str = "",
+        start_with: str = "",
+        end_with: str = "",
+    ):
+        self.replace_single_with = replace_with
+        self.replace_start_with = start_with
+        self.replace_end_with = end_with
+
+    def __call__(self, match: Match) -> str:
+        match_text = match.group(0)
+        if "<b>" in match_text or "</b>" in match_text:
+            elements = []
+            s = 0
+            element = ""
+            for match in BOLD_TAG_PATTERN.finditer(match_text):
+                if match.group(0) == "<b>":
+                    s = match.end()
+                    element = "<b>"
+                elif match.group(0) == "</b>":
+                    element = element + match_text[s:match.start()] + "</b>"
+                    elements.append(element)
+                    s = match.end()
+                    element = ""
+            if element != "":
+                element = element + match_text[s:] + "</b>"
+                elements.append(element)
+            return self.replace_start_with + " ".join(elements) + self.replace_end_with
+        else:
+            return self.replace_single_with
+
+
+class PostprocessedMatchRegexReplacerAdapter(RegexReplacerAdapter):
+    def __init__(
+            self,
+            regex_query,
+            match_postprocessor: MatchPostprocessor,
+            case_sensitive = False,
+            use_locale = False,
+            restrict_to_ascii = False,
+    ):
+        super().__init__(regex_query, match_postprocessor, case_sensitive, use_locale, restrict_to_ascii)
 
 
 class BoldPreservingRegexReplacerAdapter(RegexReplacerAdapter):
@@ -223,7 +289,7 @@ class BoldPreservingRegexReplacerAdapter(RegexReplacerAdapter):
         if not preserve_bold:
             target = replace_tag_with
         else:
-            target = _get_target_func(replace_tag_with)
+            target = PreserveHighlight(replace_tag_with)
         super().__init__(regex_query, target, case_sensitive, use_locale, restrict_to_ascii)
 
 
@@ -304,6 +370,27 @@ class URLReplacer(BoldPreservingRegexReplacerAdapter):
         super().__init__(URL_REGEX, replace_tag_with, False, False, False, preserve_bold_tag)
 
 
+class URLReplacerPreservingHighlightedEntity(PostprocessedMatchRegexReplacerAdapter):
+    """Research https://mathiasbynens.be/demo/url-regex and https://gist.github.com/dperini/729294
+    Since we need to increase recall of link-a-like entities in text, we adopt the algorithm with
+    some possible erroneous cases.
+    I used regexp from https://mathiasbynens.be/demo/url-regex by @stephenhay
+    and extended with ftps, www, wwwDDD prefixes, and "\" variant of slash ("/")"""
+
+    def __init__(
+        self,
+        replace_with: str = "<url>",
+        start_with: str = "<url>",
+        end_with: str = "</url>",
+    ) -> None:
+        match_postprocessor = PreserveHighlightedEntity(
+            replace_with=replace_with,
+            start_with=start_with,
+            end_with=end_with
+        )
+        super().__init__(URL_REGEX, match_postprocessor, False, False, False)
+
+
 class EmailReplacer(RegexReplacerAdapter):
     """Replaces all e-mail-like entities, non-latin including. Includes length check for each entity part."""
 
